/**
 ******************************************************************************
 * @file           : main.c
 * @author         : Auto-generated by STM32CubeIDE
 * @brief          : Main program body
 ******************************************************************************
 * @attention
 *
 * Copyright (c) 2022 STMicroelectronics.
 * All rights reserved.
 *
 * This software is licensed under terms that can be found in the LICENSE file
 * in the root directory of this software component.
 * If no LICENSE file comes with this software, it is provided AS-IS.
 *
 ******************************************************************************
 */

#include <stdint.h>
#include <stm32f4xx.h>
#include "FreeRTOS.h"
#include "task.h"
#include "SEGGER_SYSVIEW_Conf.h"
#include "SEGGER_SYSVIEW.h"

//#define USE_LOOP
#ifndef USE_LOOP
//#define USE_NOTIFICATION
#ifndef USE_NOTIFICATION
#define USE_INTERRUPT_NOTIFICATION
#endif
#endif

volatile unsigned int button_pressed = 0;
#ifndef USE_INTERRUPT_NOTIFICATION
TaskHandle_t buttonTaskHandler;
#endif
TaskHandle_t ledTaskHandler;

#ifdef USE_INTERRUPT_NOTIFICATION
void EXTI15_10_IRQHandler() {
  SEGGER_SYSVIEW_RecordEnterISR();
  BaseType_t aTaskWoken = pdFALSE;
  if (EXTI->PR & (1UL << 13)) {
    /* clear pending bit */
    EXTI->PR |= (1UL << 13);
    if((GPIOC->IDR & GPIO_IDR_ID13) == 0) {
        button_pressed = 1;
    } else {
        button_pressed = 0;
    }
    SEGGER_SYSVIEW_PrintfTarget("pressed=%u", button_pressed);
    xTaskGenericNotifyFromISR(
        ledTaskHandler /* xTaskToNotify */,
        0 /* uxIndexToNotify */,
        0 /* ulValue */,
        eNoAction /* eAction */,
        NULL /* pulPreviousNotificationValue */,
        &aTaskWoken /*pxHigherPriorityTaskWoken, set to NULL to not switch task immediately */
    );
    portYIELD_FROM_ISR(aTaskWoken);
  }
  SEGGER_SYSVIEW_RecordExitISR();
}
#endif

void Button_Task() {
#ifdef USE_NOTIFICATION
  volatile unsigned int button_changed = 0;
#endif
  while(1) {
    if((GPIOC->IDR & GPIO_IDR_ID13) == 0) {
#ifdef USE_LOOP
      button_pressed = 1;
#endif
#ifdef USE_NOTIFICATION
      if(!button_pressed) {
        button_pressed = 1;
        button_changed = 1;
      }
#endif
    } else {
#ifdef USE_LOOP
      button_pressed = 0;
#endif
#ifdef USE_NOTIFICATION
      if(button_pressed) {
        button_pressed = 0;
        button_changed = 1;
      }
#endif
    }

#ifdef USE_NOTIFICATION
    if(button_changed) {
      SEGGER_SYSVIEW_PrintfTarget("pressed=%u", button_pressed);
      xTaskGenericNotify(
          ledTaskHandler /* xTaskToNotify */,
          0 /* uxIndexToNotify */,
          0 /* ulValue */,
          eNoAction /* eAction */,
          NULL /* pulPreviousNotificationValue */
      );
      button_changed = 0;
    }
#endif

    //vTaskDelay(10); // delay 10 ticks may not good?
  }
}

void LED_Task() {

  while(1) {
#if defined(USE_NOTIFICATION) | defined(USE_INTERRUPT_NOTIFICATION)
    xTaskGenericNotifyWait(
      0 /* uxIndexToWaitOn */,
      0 /* ulBitsToClearOnEntry */,
      0 /* ulBitsToClearOnExit */,
      NULL /* pulNotificationValue */,
      portMAX_DELAY /* xTicksToWait*/
    );
#endif
    if(button_pressed) {
      GPIOA->BSRR |= GPIO_BSRR_BS_5;
    } else {
      GPIOA->BSRR |= GPIO_BSRR_BR_5;
    }
    //vTaskDelay(10); // delay 10 ticks may not good?
  }
}

int main(void)
{
  SystemInit();
  SEGGER_SYSVIEW_Conf();

  /* turn on clock for GPIOA, GPIOC */
  RCC->AHB1ENR |= (RCC_AHB1ENR_GPIOAEN | RCC_AHB1ENR_GPIOCEN);

  /* set PA5 to output mode */
  GPIOA->MODER &= ~GPIO_MODER_MODE5_1;
  GPIOA->MODER |=  GPIO_MODER_MODE5_0;

  /* set PC13 to input mode */
  GPIOC->MODER &= ~GPIO_MODER_MODE13_1;
  GPIOC->MODER &= ~GPIO_MODER_MODE13_0;

#ifdef USE_INTERRUPT_NOTIFICATION
  /* turn on clock for SYSCONFIG */
  RCC->APB2ENR |= RCC_APB2ENR_SYSCFGEN;
  /* select PC13 as source for EXTI */
  SYSCFG->EXTICR[3] &= ~SYSCFG_EXTICR4_EXTI13_Msk;
  SYSCFG->EXTICR[3] |= SYSCFG_EXTICR4_EXTI13_PC;
  /* enable interrupt on EXTI13 */
  EXTI->IMR |= EXTI_IMR_IM13;
  /* set up PC13 interruption on 2 edges */
  EXTI->RTSR |= EXTI_RTSR_TR13;
  EXTI->FTSR |= EXTI_FTSR_TR13;

  /* enable NVIC */
  __NVIC_SetPriority(EXTI15_10_IRQn, configLIBRARY_MAX_SYSCALL_INTERRUPT_PRIORITY+1);
  __NVIC_EnableIRQ(EXTI15_10_IRQn);
  // same as
  //NVIC->ISER[(((uint32_t)EXTI15_10_IRQn) >> 5UL)] = (uint32_t)(1UL << (((uint32_t)EXTI15_10_IRQn) & 0x1FUL));
#endif

#ifndef USE_INTERRUPT_NOTIFICATION
  xTaskCreate(
          Button_Task,
          "Button_Task",
          configMINIMAL_STACK_SIZE,
          NULL,
          1,
          &buttonTaskHandler);
#endif

  xTaskCreate(
        LED_Task,
        "LED_Task",
        configMINIMAL_STACK_SIZE,
        NULL,
        1,
        &ledTaskHandler);

  vTaskStartScheduler();
	for(;;);
}
